
#include <stdio.h>
#include <string.h>
#include "zlib.h"
#include "unzip.h"
#include "system.h"
#include "util.h"
#include "def.h"

struct _CART Cart;
BOOL         Cart_Loaded = FALSE;

// check if ROM is zipped
BOOL CheckZip(const char *name)
{
  FILE          *file;
	unsigned long  uSizeFile, uBackRead;
	unsigned long  uMaxBack  = 0xFFFF + 22;
  unsigned long  dwBuf;
  long           lReadSize, lReadPos;
  long           i;
  int            ch;

  if(!(file = fopen(name, "rb"))) return FALSE;
	if(fseek(file, 0, SEEK_END)) return fclose(file), FALSE;
	uSizeFile = ftell(file);
  if(uSizeFile < 22) return fclose(file), FALSE;
	if(uMaxBack > uSizeFile) uMaxBack = uSizeFile;
	uBackRead = 22;
  lReadSize = 4;
	while(TRUE)
	{
		lReadPos  = uSizeFile - uBackRead;		
		if(fseek(file, lReadPos, SEEK_SET)) break;
    if(fread(&dwBuf, 3, 1, file) != 1) break;
    dwBuf = dwBuf << 8;

    for(i = 3; i < lReadSize; i++)
    {
      ch = getc(file);
      if(ch == EOF) return 0;
      dwBuf = (dwBuf >> 8) | ((unsigned long)(ch & 0xFF) << 24);
      if(dwBuf == 0x06054B50) return fclose(file), TRUE; // check the magic
    }

    if     (uBackRead == uMaxBack)        break;
		else if(uBackRead + 0x400 > uMaxBack) uBackRead = uMaxBack;
		else                                  uBackRead += 0x400;

    lReadSize = ((0x0400 + 4) < uBackRead) ? (0x0400 + 4) : uBackRead;
	}

  fclose(file);

	return FALSE;
}

BOOL Cart_LoadROM(const char *name)
{
  BOOL loadOK = FALSE;
  char trueName[MAX_PATH];

  Cart_FreeROM();

  if(CheckZip(name)) // zipped ROM
  {
    unzFile       f;
    char         *ext;
    unz_file_info zipInfo;

    f = unzOpen(name);
    if(!f) return FALSE; // unable to open zip
    while(unzGetCurrentFileInfo(f, &zipInfo, trueName, MAX_PATH, (void *)0, 0, (char *)0, 0) == UNZ_OK)
    {
      ext = GetFileExtFromPath(trueName);
      if(ext)
      {
        Cart.Size = zipInfo.uncompressed_size;
        if((Cart.Size <= 0x080200 && Cart.Size >= 0x8000) &&   // check file size
           (!strcmpi(ext, ".SMS") || !strcmpi(ext, ".GG")))    // and extension
        {
          if(unzOpenCurrentFile(f) == UNZ_OK)
          {
            if(Cart.Size & 0x0200) // skip ROM header
            {
              unzReadCurrentFile(f, Cart.ROM, 0x0200);
              Cart.Size -= 0x0200;
            }
            Cart.Size = unzReadCurrentFile(f, Cart.ROM, Cart.Size);
            unzCloseCurrentFile(f);
            if(Cart.Size <= 0x080000 && Cart.Size >= 0x8000) loadOK = TRUE;
          }
        }
      }
      if(loadOK) break; // ROM load Ok
      if(unzGoToNextFile(f) != UNZ_OK) break;
    }
    unzClose(f);
  }
  else // not zipped ROM
  {
    FILE *f;
    int   c1, c2;

    f = fopen(name, "rb");
    if(!f) return FALSE; // error opening ROM

    // checking if gzip compressed file
    if(((c1 = getc(f)) == EOF) || ((c2 = getc(f)) == EOF)) return fclose(f), FALSE;

    if(c1 == 0x1F && c2 == 0x8B) // gzip compressed file
    {
      gzFile gf;
      char   tfn[MAX_PATH];

      fclose(f);
      gf = gzopen(name, "rb");
      if(!gf) return FALSE;
      // create a temporary file to hold the uncompressed data
      GetTempFilePath(tfn);
      f = fopen(tfn, "w+b");
      if(f)
      {
        Cart.Size = 0;
        // uncompressing data to temporary file
        while((c1 = gzgetc(gf)) != -1) { putc(c1, f); Cart.Size++; }
        if(Cart.Size <= 0x080200 && Cart.Size >= 0x8000)
        {
          if(!fseek(f, (Cart.Size & 0x0200) ? (Cart.Size -= 0x0200, 0x0200) : 0, SEEK_SET))
          {
            if(fread(Cart.ROM, Cart.Size, 1, f) == 1) loadOK = TRUE;
          }
        }
        fclose(f);
        remove(tfn);
      }
      gzclose(gf);
    }
    else // uncompressed ROM
    {
      if(!fseek(f, 0, SEEK_END))
      {
        Cart.Size = ftell(f);
        if(Cart.Size <= 0x080200 && Cart.Size >= 0x8000)
        {
          if(!fseek(f, (Cart.Size & 0x0200) ? (Cart.Size -= 0x0200, 0x0200) : 0, SEEK_SET))
          {
            if(fread(Cart.ROM, Cart.Size, 1, f) == 1) loadOK = TRUE;
          }
        }
      }
      fclose(f);
    }

    strncpy(trueName, name, MAX_PATH);
  }

  if(loadOK) // ROM was loaded
  {
    char *n, *e;
    if(memcmp(Cart.ROM + 0x7FF0, "TMR SEGA", 8)) return FALSE;
    // get ROM name & type
    n = GetFileNameFromPath(trueName);
    e = GetFileExtFromPath(n);
    Cart.Type = strcmpi(e, ".gg") ? CART_TYPE_SMS : CART_TYPE_GG;
    if(e) *e = 0;
    strncpy(Cart.Name, n, MAX_FILE);
    // get ROM CRC
    Cart.CRC = crc32(Cart.CRC = 0, Cart.ROM, Cart.Size);
  }
  else memset(&Cart, 0, sizeof(struct _CART));

  return (Cart_Loaded = loadOK);
}

void Cart_FreeROM(void)
{
  memset(&Cart, 0, sizeof(struct _CART));
  Cart_Loaded = FALSE;
}

BOOL Cart_LoadSRAM(const char *name)
{
  FILE *f;
  memset(Cart.SRAM, 0, 0x8000);
  f = fopen(name, "rb");
  if(!f) return FALSE;
  fread(Cart.SRAM, 1, 0x8000, f);
  fclose(f);
  return TRUE;
}


BOOL Cart_SaveSRAM(const char *name)
{
  FILE *f;
  int   i, s;
  i = 0x7FFF;
  while((i >= 0) && (Cart.SRAM[i] == 0)) i--;
  if(i < 0) return FALSE;
  i++;
  s = 1;
  while(i > s) s <<= 1;
  f = fopen(name, "wb");
  if(!f) return FALSE;
  i = fwrite(Cart.SRAM, s, 1, f);
  fclose(f);
  return (BOOL)(i == 1);
}
